from inspect import isfunction
from functools import wraps
import re
import math
import pandas as pd
import numpy as np
from scipy import interpolate
from scipy.integrate import simps
from scipy.optimize import curve_fit
from matplotlib import pyplot as plt
from xyw_utils.path import FileName


class ExcelModel:
    """
    excel文件读写模块
    """
    @staticmethod
    def read(io, sheet_name=0, header=0, names=None, usecols=None, dtype=None):
        if usecols is not None and names is not None and len(usecols) != len(names):
            raise ValueError('too many columns specified:expected %d and found %d' % (len(usecols), len(names)))
        df = pd.read_excel(io, sheet_name=sheet_name, usecols=usecols, header=header)
        if usecols is not None:
            usecols = [df.columns[item] if isinstance(item, int) else item for item in usecols]
            df = df[usecols]
        if names is not None:
            rename_dict = dict(zip(df.columns, names))
            df = df.rename(columns=rename_dict)
        if dtype is not None:
            if not isinstance(dtype, dict):
                raise TypeError('param dtype must be dict')
            for key, value in dtype.items():
                if isfunction(value):
                    df[key] = df[key].apply(value)
                elif isinstance(value, str):
                    df[key] = df[key].astype(value)
                else:
                    raise TypeError('param dtype must be dict-like of function or str')
        return df

    @staticmethod
    def write(df, io, sheet_name='Sheet1', index=False):
        df.to_excel(io, sheet_name=sheet_name, index=index)


class CsvModel:
    """
    csv文件读写模块
    """
    @staticmethod
    def read(io, header=0, names=None, usecols=None, dtype=None):
        if usecols is not None and names is not None and len(usecols) != len(names):
            raise ValueError('too many columns specified:expected %d and found %d' % (len(usecols), len(names)))
        df = pd.read_csv(io, usecols=usecols, header=header)
        if usecols is not None:
            usecols = [df.columns[item] if isinstance(item, int) else item for item in usecols]
            df = df[usecols]
        if names is not None:
            rename_dict = dict(zip(df.columns, names))
            df = df.rename(columns=rename_dict)
        if dtype is not None:
            if not isinstance(dtype, dict):
                raise TypeError('param dtype must be dict')
            for key, value in dtype.items():
                if isfunction(value):
                    df[key] = df[key].apply(value)
                elif isinstance(value, str):
                    df[key] = df[key].astype(value)
                else:
                    raise TypeError('param dtype must be dict-like of function or str')
        return df

    @staticmethod
    def write(df, io, index=False):
        df.to_csv(io, index=index)


class FileProcessorFacade:
    """
    文件读写的外观类
    """
    def __init__(self):
        self.__excel_model = ExcelModel()
        self.__csv_model = CsvModel()

    def read(self, io, sheet_name=0, header=0, names=None, usecols=None, dtype=None):
        if not isinstance(io, str):
            raise TypeError('param io must be string')
        file = FileName(io)
        if file.is_ext(['.xlsx', '.xls']):
            return self.__excel_model.read(
                file.fullpath, sheet_name=sheet_name, header=header,
                names=names, usecols=usecols, dtype=dtype,
            )
        elif file.is_ext('.csv'):
            return self.__csv_model.read(
                file.fullpath, header=header, names=names,
                usecols=usecols, dtype=dtype,
            )
        else:
            raise ValueError('unsupport file type')

    def write(self, df, io, sheet_name='Sheet1', index=False):
        file = FileName(io)
        if file.is_ext(['.xlsx', '.xls']):
            file.make_dir()
            self.__excel_model.write(df, file.fullpath, sheet_name=sheet_name, index=index)
        elif file.is_ext('.csv'):
            file.make_dir()
            self.__csv_model.write(df, file.fullpath, index=index)
        else:
            raise ValueError('unsupport file type')


def validate_df(func):
    """
    判断参数是否为pandas中DataFrame类型的装饰器
    :param func:
    :return:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if isinstance(args[0], pd.DataFrame):
            # 将第一个df参数替换为其复制
            args = list(args)
            args[0] = args[0].copy()
            return func(*args, **kwargs)
        else:
            raise TypeError('support pandas.DataFrame only')
    return wrapper


def validate_dfs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        args = list(args)
        if isinstance(args[0], list):
            for item in args[0]:
                if not isinstance(item, pd.DataFrame):
                    raise TypeError('support list of pandas.DataFrame only')
        else:
            raise TypeError('support list of pandas.DataFrame only')
        args[0] = [item.copy() for item in args[0]]
        return func(*args, **kwargs)
    return wrapper


@validate_df
def remove_unnamed(df):
    """
    移除df中的Unnamed列
    :param df:
    :return:
    """
    pattern = re.compile(r'^Unnamed: \d*$')
    unnamed = [item for item in df.columns if pattern.match(item)]
    return df.drop(columns=unnamed)


@validate_df
def remove_nan(df):
    """
    移除df中有无效值的行
    :param df:
    :return:
    """
    return df.dropna(axis=0, how='any').reset_index(drop=True)


@validate_df
def interpolate_nan(df, kind='linear', drop=True):
    """
    对无效值进行插值处理
    :param df:
    :param kind: 插值方法
    :param drop: 是否去除插值后仍未nan的首尾数据行
    :return:
    """
    for x in range(df.shape[1] - 1):
        tem = df.iloc[:, x + 1]
        # 查找非NaN数据索引，此处必须用isnan方法进行判定
        tem = tem.index[np.where(np.isnan(tem) == False)[0]]
        # 调用插值函数，获取插值表达式
        f = interpolate.interp1d(df.iloc[tem, 0], df.iloc[tem, x + 1], kind=kind)
        # 为无效值赋值
        df.iloc[tem[0]:tem[-1], x + 1] = f(df.iloc[tem[0]:tem[-1], 0])
    if drop:
        df = df.dropna(how='any')
    return df.reset_index(drop=True)


@validate_df
def init_split(df):
    """
    单独时间列df数据的初始化
    :param df:
    :return:
    """
    df = remove_unnamed(df)
    if df.shape[1] % 2:
        raise RuntimeError('columns of the df is not an even number')
    pattern = re.compile(r'^(.*)\.\d*$')
    res = []
    for i in range(df.shape[1] // 2):
        tem = df.iloc[:, [i*2, i*2+1]]
        tem = remove_nan(tem)
        match_res = pattern.match(tem.columns[0])
        if match_res:
            tem.rename(columns={tem.columns[0]: match_res.group(1)}, inplace=True)
        res.append(tem)
    return res


@validate_df
def init_merged(df):
    """
    唯一时间列df数据的初始化
    :param df:
    :return:
    """
    df = remove_unnamed(df)
    res = []
    for i in range(df.shape[1] - 1):
        tem = df.iloc[:, [0, i+1]]
        tem = remove_nan(tem)
        res.append(tem)
    return res


@validate_df
def set_zero(df):
    """
    将df数据的第一列起始值归零
    :param df:
    :return:
    """
    df.iloc[:, 0] = df.iloc[:, 0] - df.iloc[0, 0]
    return df


@validate_df
def resample_single(df, step=0.01, kind=None):
    """
    重采样
    :param df: 两列的df数据，第一列一般为时间列
    :param step: 采样步长
    :param kind: 插值方法
    :return:
    """
    if kind is None:
        kind = 'linear'
    if df.shape[1] != 2:
        raise RuntimeError('the number of columns must be two')
    f = interpolate.interp1d(df.iloc[:, 0], df.iloc[:, 1], kind=kind)
    start = math.ceil(df.iloc[0, 0] / step)
    end = math.floor(df.iloc[-1, 0] / step)
    x = np.arange(start, end + 1, step=1, dtype=int)
    x = x * step
    y = f(x)
    return pd.DataFrame(np.vstack((x, y)).T, columns=df.columns)


@validate_df
def get_derivation(df, method='grad'):
    """
    离散数据求导
    :param df:
    :param method: 微分方法，diff方法下第一行数据为NAN
    :return:
    """
    x = df.iloc[:, 0]
    y = df.iloc[:, 1]
    # 中心差分法
    if method == 'grad':
        dx = np.gradient(x)
        dy = np.gradient(y)
    # 向前差分法
    elif method == 'diff':
        dx = x.diff()
        dy = y.diff()
    else:
        raise ValueError('not support such method, grad and diff only')
    res = dy / dx
    name = method + '[' + x.name + ',' + y.name + ']'
    res = pd.Series(res, name=name)
    return res


@validate_df
def get_integral(df, method='trapz', last_only=False):
    """
    离散数据求积分
    :param df:
    :param method: 积分方法
    :param last_only: 是否仅返回最终积分结果
    :return:
    """
    # 提取x，y
    x = df.iloc[:, 0]
    y = df.iloc[:, 1]
    # 获取x，y列名称
    x_name = x.name
    y_name = y.name

    # 辛普森法
    if method == 'simps':
        if last_only:
            return simps(y, x)
        res = [simps(y[0:i+1], x[0:i+1]) for i in range(len(x))]
    # 复杂梯形法
    elif method == 'trapz':
        if last_only:
            return np.trapz(y, x)
        res = [np.trapz(y[0:i+1], x[0:i+1]) for i in range(len(x))]
    else:
        raise ValueError('not support such method, simps and trapz only')

    name = method + '[' + x_name + ',' + y_name + ']'
    res = pd.Series(res, name=name)

    return res


@validate_dfs
def resample_all(dfs, step=0.01, kind=None):
    """
    将列表中的df统一重采样
    :param dfs:
    :param step:
    :param kind:
    :return:
    """
    return [resample_single(item, step=step, kind=kind) for item in dfs]


@validate_dfs
def merge_all(dfs, drop_nan=False, zero=False):
    """
    将列表中的多个df拼接为一个df
    :param drop_nan: 是否删除有nan的行
    :param dfs:
    :param zero: 是否将第一列初始值归零
    :return:
    """
    res = dfs[0]
    name = res.columns[0]
    for i in range(len(dfs)-1):
        res = pd.merge(res, dfs[i+1], how='outer', on=name)
    if drop_nan:
        res = remove_nan(res)
    res = res.sort_values(by=name, ascending=True).reset_index(drop=True)
    if zero:
        res = set_zero(res)
    return res


@validate_df
def get_local_outlier_factor(df, k):
    """
    计算局部离群因子值
    :param df:
    :param k: 取离样本点p第k个最近的距离
    :return:
    """
    # def f1(x, df, k):
    #     tem = np.sqrt(((df - x) ** 2).sum(axis=1))
    #     tem = tem.sort_values()
    #     return 1 / tem.iat[k], tem.iloc[1:k + 1].index
    #
    # def f2(x, df):
    #     tem = df[0][x[1]].mean()
    #     return tem / x[0]

    if not isinstance(k, int):
        raise TypeError('param k must be int')
    if len(df) < k+1:
        raise RuntimeError('param k must be smaller than the length of df')

    name = 'lof' + str(list(df.columns))

    # df = df.reset_index(drop=True)
    # df = df.apply(f1, axis=1, args=(df, k)).apply(pd.Series)
    # res = df.apply(f2, axis=1, args=(df,))
    # res.name = name
    # return res

    rho = []
    index = []
    for i in range(len(df)):
        tem = np.sqrt(((df - df.iloc[i, :])**2).sum(axis=1))
        tem = tem.sort_values()
        rho.append(1 / tem.iat[k])
        index.append(tem.iloc[1:k+1].index)
    rho = pd.Series(rho)
    res = []
    for i in range(len(df)):
        tem = rho[index[i]].mean()
        res.append(tem / rho[i])
    res = pd.Series(res, name=name)
    return res


@validate_df
def remove_outlier(df, contamination=0.01, k=20):
    """
    移除数据中离群点所在行
    :param df:
    :param contamination: 离群点比例
    :param k: 局部离群因子算法中的k
    :return:
    """
    tem = get_local_outlier_factor(df, k=k)
    tem = tem.sort_values(ascending=False)
    contamination = math.ceil(len(df)*contamination)
    index = list(tem.iloc[0:contamination].index)
    df.iloc[index, :] = np.nan
    df = remove_nan(df)
    return df


@validate_df
def get_linear_fitting(df, method):
    """
    常用最小二乘法线性拟合
    :param df:
    :param method:
    :return:
    """
    # 无常数项一次拟合
    def fun1(x, a):
        return a * x

    # 无常数项二次拟合
    def fun2(x, a, b):
        return a * x ** 2 + b * x

    # 无常数项三次拟合
    def fun3(x, a, b, c):
        return a * x ** 3 + b * x ** 2 + c * x

    x = df.iloc[:, 0]
    y = df.iloc[:, 1]

    if method in [1, 2, 3]:
        popt = np.polyfit(x, y, method)
    elif method == 4:
        popt, _ = curve_fit(fun1, x, y)
        popt = np.append(popt, 0)
    elif method == 5:
        popt, _ = curve_fit(fun2, x, y)
        popt = np.append(popt, 0)
    elif method == 6:
        popt, _ = curve_fit(fun3, x, y)
        popt = np.append(popt, 0)
    else:
        raise ValueError('param method must be int between 1 and 6')
    return np.poly1d(popt)


@validate_df
def plot(df):
    p = plt.plot(df.iloc[:, 0], df.iloc[:, 1])
    plt.show()
    return p


if __name__ == '__main__':
    # from io import StringIO, BytesIO
    # import os
    #
    # data = ('col1,col2,col3\n'
    #         'a,b,1\n'
    #         'a,b,2\n'
    #         'c,d,3')
    #
    # usercols = [2, 0]
    # names = ['x', 'y']
    # header = None
    #
    # reader = CsvModel()
    # reader2 = FileProcessorFacade()
    # df = reader.read(StringIO(data), usecols=[2, 1, 0], names=['x', 'y', 'z'])
    # # df = reader.read(StringIO(data), usecols=[1, 0], names=['x', 'y'], dtype={'x': lambda x: x == 'b'})
    #
    # # df = pd.read_csv(StringIO(data), usecols=[1, 0], names=['x', 'y'], header=1, dtype={'x': bool})
    # print(df)
    # print(df.dtypes)
    #
    # reader2.write(df, os.path.join('xue', 'x.csv'))
    processor = FileProcessorFacade()
    # df = processor.read(r'C:\Users\Administrator\Desktop\JS6108GHBEV31\65%载荷\ZZ-CCBC-65%-1.csv', sheet_name=2)
    df = processor.read(r'C:\Users\Administrator\Desktop\cw\wthz-l-1.xlsx', sheet_name=2)
    # df = processor.read(r'1.xlsx', sheet_name=2)
    dfs = init_merged(df)
    df = resample_single(dfs[0], step=0.01)
    re = get_integral(df)
    # df2 = resample_all(dfs, step=0.1)
    # df2 = merge_all(df2)
    # df = init_split(df)
    print(df)



import numpy as np


def rad_to_degree(rad):
    return rad / np.pi * 180


def degree_to_rad(degree):
    return degree / 180 * np.pi


def sind(degree):
    return np.sin(degree_to_rad(degree))


def cosd(degree):
    return np.cos(degree_to_rad(degree))



import pygal
from pygal.style import DefaultStyle, LightColorizedStyle, DarkStyle
from xyw_utils.data import *
from xyw_utils.math import sind, cosd


def driving_resistance(gvw, vehicle_type='coach'):
    """
    根据车辆总质量和车型，按照GB/T 18386-2017附录A中表格估算车辆行驶阻力系数
    :param gvw: 车辆最大总质量(kg)
    :param vehicle_type: 车辆类型，公路客车或城市客车，coach or bus
    :return: F = A + B * V + C * V^2,F(N),V(km/h)
    """
    m = np.array([3500, 4500, 5500, 7000, 8500, 10500, 12500, 14500, 16500, 18000, 22000, 25000])
    a1 = np.array([450.9, 481.0, 511.0, 556.1, 601.1, 661.2, 721.3, 781.4, 841.5, 886.5, 1006.7, 1096.8])
    b1 = np.array([2.29, 2.66, 3.02, 3.57, 4.12, 4.85, 5.58, 6.32, 7.05, 7.60, 9.06, 10.16])
    c1 = np.array([0.115, 0.119, 0.123, 0.129, 0.134, 0.142, 0.150, 0.158, 0.165, 0.171, 0.187, 0.198])
    a2 = np.array([432.9, 473.2, 513.6, 574.1, 634.6, 715.2, 795.9, 876.6, 957.3, 1017.8, 1179.1, 1300.1])
    b2 = np.array([2.67, 2.79, 2.91, 3.10, 3.28, 3.53, 3.78, 4.02, 4.27, 4.46, 4.95, 5.32])
    c2 = np.array([0.113, 0.120, 0.127, 0.138, 0.148, 0.162, 0.176, 0.190, 0.204, 0.214, 0.242, 0.263])

    if vehicle_type == 'coach':  # 公路车
        f = interpolate.interp1d(m, a1, kind='linear')
        a = f(gvw)
        f = interpolate.interp1d(m, b1, kind='linear')
        b = f(gvw)
        f = interpolate.interp1d(m, c1, kind='linear')
        c = f(gvw)
    elif vehicle_type == 'bus':  # 城市客车，即公交车
        f = interpolate.interp1d(m, a2, kind='linear')
        a = f(gvw)
        f = interpolate.interp1d(m, b2, kind='linear')
        b = f(gvw)
        f = interpolate.interp1d(m, c2, kind='linear')
        c = f(gvw)
    else:
        raise ValueError('not support such vehicle type,coach and bus only')

    p = [c, b, a]
    p = np.poly1d(p)
    return p


@validate_df
def steady_static_circular_test(df, l, t, a_y, v_x, v_yaw, roll):
    """
    稳态回转试验数据处理，主要参考：
    《GB/T 6323-2014 汽车操纵稳定性试验方法》10.4
    :param df: 包含所需数据的df（已筛选）
    :param l: 轴距，单位m
    :param t: 时间列名称，单位s
    :param a_y: 侧向加速度列名称，单位m/s^2
    :param v_x: 前进速度列名称，单位km/h
    :param v_yaw: 横摆角速度列名称，单位°/s
    :param roll: 车身侧倾角列名称，单位°
    :return:
    """
    # 提取并重命名相关数据列
    df = df[[t, a_y, v_x, v_yaw, roll]]
    df = df.rename(columns={t: 't', a_y: 'a_y', v_x: 'v_x', v_yaw: 'v_yaw', roll: 'roll'})

    # 计算侧向加速度修正值与转弯半径
    g = 9.81
    df['a_y_true'] = (df.a_y - g * sind(df.roll)) / cosd(df.roll)
    df['r'] = 180 / math.pi * df.v_x / 3.6 / df.v_yaw

    # 计算初始半径
    p_r = get_linear_fitting(df[['a_y_true', 'r']], method=3)
    r0 = p_r(0)

    # 根据初始半径正负判断转弯方向
    if r0 < 0:
        df['r'] = np.abs(df.r)
        r0 = np.abs(r0)
        direction = 'left'
    else:
        direction = 'right'

    # 计算转弯半径比
    df['ratio_r'] = df.r / r0

    # 计算前后轴侧偏角差
    df['d_sideslip_angle'] = 180 / math.pi * l * (1 / r0 - 1 / df.r)

    # 计算侧向加速度与转弯半径比拟合曲线
    p_ratio = get_linear_fitting(df[['a_y_true', 'ratio_r']], method=3)
    # 计算侧向加速度与前后轴侧偏角差拟合曲线
    p_d = get_linear_fitting(df[['a_y_true', 'd_sideslip_angle']], method=6)
    # 计算侧向加速度与车身侧倾角拟合曲线
    # 此处考虑到车身有起始侧倾角，因此没有使用无常数项的二次拟合，而是使用普通二次拟合
    p_roll = get_linear_fitting(df[['a_y_true', 'roll']], method=2)

    # 计算中性转向点
    a_neutral = np.roots(p_d.deriv())
    a_neutral = -a_neutral if direction == 'left' else a_neutral
    # 计算不足转向度
    degree_understeer = p_d(-2 if direction == 'left' else 2) / 2
    # 计算车身侧倾度
    roll_rate_body = (p_roll(-2 if direction == 'left' else 2) - p_roll(0)) / 2

    # 绘图时标题中的后缀
    suffix = '（' + ('左转' if direction == 'left' else '右转') + '）'

    # 绘图时用到点的x坐标
    x_max = df.a_y_true.max()
    x_min = df.a_y_true.min()
    x_range = np.arange(0, x_max, 0.01) if x_max > 0 else np.arange(x_min, 0, 0.01)

    # 绘制转弯半径比特性曲线
    ratio_chart = pygal.XY(
        legend_at_bottom=True,
        legend_at_bottom_columns=2,
        style=DefaultStyle(font_family='SimHei'),
        x_title='侧向加速度 a_y(m/s²)',
        y_title='转弯半径比 R/R_0',
    )
    ratio_chart.title = '转弯半径比特性曲线' + suffix
    ratio_chart.add('实际数据点', list(zip(df.a_y_true, df.ratio_r)), stroke=False)
    ratio_chart.add(
        'y={0:.4g}x³+{1:.4g}x²+{2:.4g}x+{3:.4g}'.format(*tuple(p_ratio.c)).replace('+-', '-'),
        [(x, p_ratio(x)) for x in x_range],
        show_dots=False, stroke=True, stroke_style={'width': 3}
    )
    # ratio_chart.render_in_browser()

    # 绘制前后轴侧偏角差曲线
    d_chart = pygal.XY(
        legend_at_bottom=True,
        legend_at_bottom_columns=2,
        style=DefaultStyle(font_family='SimHei'),
        x_title='侧向加速度 a_y(m/s²)',
        y_title='前后轴侧偏角差 δ_1-δ_2(°)',
    )
    d_chart.title = '前后轴侧偏角差特性曲线' + suffix
    d_chart.add('实际数据点', list(zip(df.a_y_true, df.d_sideslip_angle)), stroke=False)
    d_chart.add(
        'y={0:.4g}x³+{1:.4g}x²+{2:.4g}x'.format(*tuple(p_d.c)).replace('+-', '-'),
        [(x, p_d(x)) for x in x_range],
        show_dots=False, stroke=True, stroke_style={'width': 3}
    )
    # d_chart.render_in_browser()

    # 绘制车身侧倾角曲线
    roll_chart = pygal.XY(
        legend_at_bottom=True,
        legend_at_bottom_columns=2,
        style=DefaultStyle(font_family='SimHei'),
        x_title='侧向加速度 a_y(m/s²)',
        y_title='前后轴侧偏角差 Ψ(°)',
    )
    roll_chart.title = '车身侧倾角特性曲线' + suffix
    roll_chart.add('实际数据点', list(zip(df.a_y_true, df.roll)), stroke=False)
    roll_chart.add(
        'y={0:.4g}x²+{1:.4g}x+{2:.4g}'.format(*tuple(p_roll.c)).replace('+-', '-'),
        [(x, p_roll(x)) for x in x_range],
        show_dots=False, stroke=True, stroke_style={'width': 3}
    )
    # roll_chart.render_in_browser()

    return {
        'df': df,
        'charts': (ratio_chart, d_chart, roll_chart),
        'curves': (p_ratio, p_d, p_roll),
        'results': (direction, r0, a_neutral, degree_understeer, roll_rate_body),
    }


@validate_df
def returnability_test(df, leave_time, t, v_yaw, angle, precision=0.1):
    """
    低速转向回正试验数据处理，主要参考：
    《GB/T 6323-2014 汽车操纵稳定性试验方法》8.4
    :param df: 包含所需数据的df
    :param leave_time: 松手时刻
    :param t: 时间列名称，单位s
    :param v_yaw: 横摆角速度列名称，单位°/s
    :param angle: 转向盘转角列名称，单位°
    :param precision: 筛选精度
    :return:
    """
    # 提取并重命名相关数据列
    df = df[[t, v_yaw, angle]]
    df = df.rename(columns={t: 't', v_yaw: 'v_yaw', angle: 'angle'})

    # 截取松手至三秒后时刻数据段
    df_true = df.loc[(df.t >= leave_time) & (df.t <= leave_time + 3)]
    if len(df_true) < 2:
        raise RuntimeError('leave time out of range')
    if abs(df_true.iloc[0].t - leave_time) > precision:
        raise RuntimeError('there is no leave time satisfying %f precision in the data' % precision)
    if abs(df_true.iloc[-1].t - leave_time - 3) > precision:
        raise RuntimeError('there is no time three seconds later than the leave time'
                           ' satisfying %f precision in the data' % precision)

    # 计算横摆角速度初始值和结束值
    start_yaw = df_true.iloc[0].v_yaw
    end_yaw = df_true.iloc[-1].v_yaw  # 残余横摆角速度

    # 根据起始横摆角速度判断转弯方向
    direction = 'left' if start_yaw < 0 else 'right'

    # 计算横摆角速度总方差
    variance = (((df_true.v_yaw / start_yaw) ** 2).sum() - 0.5) * 3 / len(df_true)

    # 绘图时标题中的后缀
    suffix = '（' + ('左转' if direction == 'left' else '右转') + '）'

    # 截取绘图所用数据
    df_chart = df.loc[(df.t > leave_time - 3) & (df.t < leave_time + 5)]
    init_time = df_chart.iloc[0].t
    df_chart = set_zero(df_chart)

    # 绘制低速转向回正横摆角速度时域曲线
    yaw_chart = pygal.XY(
        legend_at_bottom_columns=4,
        style=DefaultStyle(font_family='SimHei'),
        x_title='时间 t(s)',
        y_title='横摆角速度 Δr(°/s)',
    )
    yaw_chart.title = '低速转向回正横摆角速度时域曲线' + suffix
    yaw_chart.add('横摆角速度', list(zip(df_chart.t, df_chart.v_yaw)),
                  stroke=True, show_dots=False, stroke_style={'width': 3})
    yaw_chart.add('松手时刻', [(df_true.iloc[0].t - init_time, df_true.iloc[0].v_yaw)], dots_size=5)
    yaw_chart.add('3s时刻', [(df_true.iloc[-1].t - init_time, df_true.iloc[-1].v_yaw)], dots_size=5)
    # yaw_chart.render_in_browser()

    # 绘制低速转向回正转向盘转角时域曲线
    angle_chart = pygal.XY(
        legend_at_bottom_columns=4,
        style=DefaultStyle(font_family='SimHei'),
        x_title='时间 t(s)',
        y_title='转向盘转角 δ(°)',
    )
    angle_chart.title = '低速转向回正转向盘转角时域曲线' + suffix
    angle_chart.add('转向盘转角', list(zip(df_chart.t, df_chart.angle)),
                    stroke=True, show_dots=False, stroke_style={'width': 3})
    angle_chart.add('松手时刻', [(df_true.iloc[0].t - init_time, df_true.iloc[0].angle)], dots_size=5)
    angle_chart.add('3s时刻', [(df_true.iloc[-1].t - init_time, df_true.iloc[-1].angle)], dots_size=5)
    # angle_chart.render_in_browser()

    return {
        'charts': (yaw_chart, angle_chart),
        'results': (end_yaw, variance),
    }


@validate_df
def steering_efforts_test(df, d, t, v_x, angle, torque):
    """
    转向轻便性试验数据处理，主要参考：
    《GB/T 6323-2014 汽车操纵稳定性试验方法》9.4
    :param df: 筛选截取好的一周双纽线数据
    :param d: 方向盘直径，单位m
    :param t: 时间列名称，单位s
    :param v_x: 前进车速，单位km/h
    :param angle: 转向盘转角，单位°
    :param torque: 转向盘力矩，单位N*m
    :return:
    """
    # 提取并重命名相关数据列
    df = df[[t, v_x, angle, torque]]
    df = df.rename(columns={t: 't', v_x: 'v_x', angle: 'angle', torque: 'torque'})

    # 计算平均车速
    v_average = df.v_x.mean()

    # 计算转向盘最大作用力矩
    m_max = df.torque.abs().max()

    # 计算左右转最大转角
    angle_left_max = df.angle.min()
    angle_right_max = df.angle.max()
    if angle_left_max >= 0:
        raise RuntimeError('error original data,the angle of turning left should be negative')
    if angle_right_max <= 0:
        raise RuntimeError('error original data,the angle of turning right should be positive')

    # 计算转向盘最大作用力
    f_max = m_max * 2 / d

    # 计算绕双纽线路径一周的转向盘作用功
    m = df.torque.iloc[0:len(df) - 1].reset_index(drop=True)
    s1 = df.angle.iloc[1:len(df)].reset_index(drop=True)
    s2 = df.angle.iloc[0:len(df) - 1].reset_index(drop=True)
    w = np.pi / 180 * ((s1 - s2) * m).sum()

    # 计算转向盘平均摩擦力矩
    m_swf = 180 / np.pi * w / 2 / (abs(angle_left_max) + abs(angle_right_max))

    # 计算转向盘平均摩擦力
    f_swf = 2 * m_swf / d

    # 绘制转向盘力矩转角曲线
    chart = pygal.XY(
        style=DefaultStyle(font_family='SimHei'),
        x_title='转向盘转角 δ(°)',
        y_title='转向力矩 M(N·m)',
        title='转向盘力矩转角曲线',
        show_legend=False,
    )
    chart.add('实际数据点', list(zip(df.angle, df.torque)))
    # chart.render_in_browser()

    return {
        'results': (v_average, angle_left_max, angle_right_max,
                    m_max, f_max, w, m_swf, f_swf),
        'chart': chart,
    }


def draw_lemniscate(radius, width=3.0):
    """
    绘制双纽线摆桩图，主要参考：
    《GB/T 6323-2014 汽车操纵稳定性试验方法》9.3.1
    :param radius: 试验汽车前外轮的最小转弯半径，单位m
    :param width: 车宽，单位m
    :return:
    """
    angle = np.arange(0, 361, 0.1)
    r_min = radius * 1.1  # 双纽线最小曲率半径
    d = r_min * 3  # x轴最大值

    # 计算绘制双纽线所需xy坐标点
    l = d * np.sqrt(np.array([0 if item < 0 else item for item in cosd(2 * angle)]))
    x = l * cosd(angle)
    y = l * sind(angle)

    # 计算双纽线y轴最高点坐标
    x_max = d * np.sqrt(cosd(30 * 2)) * cosd(30)
    y_max = d * np.sqrt(cosd(30 * 2)) * sind(30)

    # 计算摆桩的偏移量
    offset = width / 2 + 0.5  # 车宽的一半加50cm

    # 标桩点
    stakes = [
        (d-offset, 0), (d+offset, 0), (x_max, y_max-offset), (x_max, y_max+offset),
        (x_max, -y_max-offset), (x_max, -y_max+offset), (offset, 0)
    ]
    stakes.extend([(-(item[0]), item[1]) for item in stakes])
    stakes.extend([(0, -offset), (0, offset)])

    # 绘制摆桩示意图
    chart = pygal.XY(style=DefaultStyle(font_family='SimHei'))
    chart.title = '双纽线摆桩示意图（最小转弯半径{0}m，车宽{1}m）'.format(radius, width)
    chart.add('双纽线', list(zip(x, y)), show_dots=False)
    chart.add('标桩点', stakes, stroke=False, formatter=lambda item: '({0:.4f},{1:.4f})'.format(*item), dots_size=5)
    chart.add('x={0:.4f}'.format(offset), [(offset, y_max+offset), (offset, -y_max-offset)], show_dots=False)
    chart.add(
        'x={0:.4f}'.format(x_max),
        [(x_max, y_max + offset), (x_max, -y_max - offset)],
        show_dots=False,
    )
    chart.add(
        'x={0:.4f}'.format(d - offset),
        [(d - offset, y_max + offset), (d - offset, -y_max - offset)],
        show_dots=False,
    )
    chart.add(
        'x={0:.4f}'.format(d + offset),
        [(d + offset, y_max + offset), (d + offset, -y_max - offset)],
        show_dots=False,
    )
    chart.add(
        'y={0:.4f}'.format(offset),
        [(-d - offset, offset), (d + offset, offset)],
        show_dots=False,
    )
    chart.add(
        'y={0:.4f}'.format(y_max - offset),
        [(-d - offset, y_max - offset), (d + offset, y_max - offset)],
        show_dots=False,
    )
    chart.add(
        'y={0:.4f}'.format(y_max + offset),
        [(-d - offset, y_max + offset), (d + offset, y_max + offset)],
        show_dots=False,
    )
    # chart.render_in_browser()
    # chart.render_to_png('re.png')
    # chart.render_to_file('re.svg')

    return {
        'chart': chart,
        'results': (stakes, r_min, d, offset, x_max, y_max),
    }


@validate_df
def energy_consumption_test(df, factor, t, current, voltage, last_only=True):
    """
    根据电压电流和时间，计算能耗
    :param df:
    :param factor: 计算功率时乘的系数
    :param t: 时间列名称，单位s
    :param current: 电流列名称，单位A
    :param voltage: 电压列名称，单位V
    :param last_only: 是否只返回结果
    :return:
    """
    # 提取并重命名相关数据列
    df = df[[t, current, voltage]]
    df = df.rename(columns={t: 't', current: 'current', voltage: 'voltage'})

    # 计算功率和能耗
    df['power'] = df.current * df.voltage * factor

    return get_integral(df[['t', 'power']], last_only=last_only) / 3600000


@validate_df
def filter_by_fixed_interval(df, column, start, end, step=5, precision=0.2):
    """
    按固定步长，筛选数据中的部分数据。常用于动力性等试验中提取以5为间隔的车速点。
    筛选规则如下：由上往下依次遍历df，目标列数据首次超过目标点时，比较该点与上一点与目标点之间的距离，取较小的作为结果
    :param df:
    :param column: 用于筛选的数据列
    :param start: 起始数值
    :param end: 结束数值
    :param step: 步长，必须能被start-end整除
    :param precision: 数据点精度
    :return:
    """
    # 参数判断
    # if not(isinstance(start, int) and isinstance(end, int) and isinstance(step, int)):
    #     raise TypeError('param start,end and step must be integer')
    if start == end:
        raise ValueError('param start and end cannot be equal')
    if (end - start) % step:
        raise ValueError('param step cannot be divisible')

    # 截取起始数据段
    df = df.reset_index(drop=False)
    start_index = df.loc[(df[column] - start).abs().argsort()].index[0]
    end_index = df.loc[(df[column] - end).abs().argsort()].index[0]
    df = df.loc[start_index: end_index]
    df = df.reset_index(drop=True)

    # 判断起始数据精度是否达标
    if abs(df[column].iloc[0] - start) > precision:
        raise RuntimeError('there is no column satisfying the precision of start')
    if abs(df[column].iloc[-1] - end) > precision:
        raise RuntimeError('there is no column satisfying the precision of end')

    # 查找各目标点
    res = df.iloc[[0, -1]]
    length = len(df) - 2
    pointer = 0
    step = abs(step) if start < end else -abs(step)

    for i in range(abs(round((end - start) / step))):
        target = (i + 1) * step + start
        while True:
            pointer = pointer + 1
            if pointer > length:
                break
            if (start < end and df[column][pointer] > target)\
                    or (start > end and df[column][pointer] < target):
                before = abs(df[column][pointer - 1] - target)
                after = abs(df[column][pointer] - target)
                if before > precision and after > precision:
                    raise RuntimeError('{} cannot satisfy the precision'.format(target))
                if after < before:
                    res = res.append(df.loc[pointer])
                else:
                    res = res.append(df.loc[pointer - 1])
                break

    # 判断目标点是否全部找到
    if len(res) != (abs(round((end - start) / step)) + 1):
        raise RuntimeError('can not find all target')

    res = res.sort_values(by=column)
    res = res.set_index(res['index'].values).drop(columns=['index'])
    res.index = res.index.astype(int)

    return res


@validate_df
def brake_test(df, t, v_x, s=None):
    """
    计算制动试验结果，主要参考：
    《GB 12676-2014 商用车辆和挂车制动系统技术要求及试验方法》 5.1.1.3
    :param df:
    :param t: 时间列名称
    :param v_x: 前进车速列名称
    :param s: 行驶距离列名称，为None则自动根据时间车速积分
    :return:
    """
    # 提取并重命名相关数据列
    if s is None:
        df = df[[t, v_x]]
        df = df.rename(columns={t: 't', v_x: 'v_x'})
        df['s'] = get_integral(df[['t', 'v_x']]) / 3.6
    else:
        df = df[[t, v_x, s]]
        df = df.rename(columns={t: 't', v_x: 'v_x', s: 's'})

    # 计算起始车速和结束车速
    start_speed = df.v_x.iloc[0]
    # end_speed = df.v_x.iloc[-1]

    # 计算起始距离和结束距离
    start_s = df.s.iloc[0]
    end_s = df.s.iloc[-1]

    f = interpolate.interp1d(df.v_x, df.s)
    vb = 0.8 * start_speed
    ve = 0.1 * start_speed
    sb = f(vb) - start_s
    se = f(ve) - start_s

    # 计算充分发出的平均减速度
    mfdd = (vb**2 - ve**2) / 25.92 / (se - sb)

    # 计算制动时间和制动距离
    brake_t = df.t.iloc[-1] - df.t.iloc[0]
    brake_s = end_s - start_s

    return mfdd, start_speed, brake_t, brake_s


if __name__ == '__main__':
    # p = driving_resistance(12500)
    # print(p)
    res = draw_lemniscate(15.94/2, width=2.2)
    print()



import img2pdf
from pikepdf import Pdf
from xyw_utils.path import FileName


A3 = (img2pdf.mm_to_pt(297), img2pdf.mm_to_pt(420))
A4 = (img2pdf.mm_to_pt(210), img2pdf.mm_to_pt(297))
A5 = (img2pdf.mm_to_pt(148), img2pdf.mm_to_pt(210))
B4 = (img2pdf.mm_to_pt(257), img2pdf.mm_to_pt(364))
B5 = (img2pdf.mm_to_pt(182), img2pdf.mm_to_pt(257))


def img_to_pdf(imgs, pdf_path, page_size=None, reverse=False):
    """
    转换图片为pdf格式并合并，为了保证无损转换，不支持有alpha通道的图片
    :param imgs: 图片地址列表，需要按合并顺序排列
    :param pdf_path: pdf文件的保存地址
    :param page_size: pdf文件的页面尺寸，默认为A4
    :param reverse: 是否反转页面尺寸，即长宽尺寸互换
    :return:
    """
    # 检查保存地址是否为.pdf后缀
    pdf = FileName(pdf_path)
    if not pdf.is_ext('.pdf'):
        raise ValueError('type of saving file must be pdf')

    # 检查imgs参数
    if isinstance(imgs, str):
        imgs = [imgs]
    if isinstance(imgs, list):
        for img in imgs:
            if not FileName(img).is_ext(['.png', '.jpeg', '.jpg']):
                raise ValueError('\'{}\' is not image'.format(img))
    else:
        raise TypeError('param imgs must be str or list of str')

    # 默认页面尺寸A4
    if page_size is None:
        page_size = A4

    # 长宽反转
    if reverse:
        page_size = page_size[::-1]
    layout_fun = img2pdf.get_layout_fun(page_size)

    # 转换图片为pdf格式并合并到指定路径
    pdf.make_dir()
    with open(pdf.fullpath, 'wb') as f:
        f.write(img2pdf.convert(imgs, layout_fun=layout_fun))


def merge_pdfs(pdfs, pdf_path):
    """
    合并多个pdf文件
    :param pdfs: 待合并pdf文件地址列表
    :param pdf_path: 最终pdf文件保存地址
    :return: pikepdf.Pdf
    """
    # 检查保存地址是否为.pdf后缀
    pdf = FileName(pdf_path)
    if not pdf.is_ext('.pdf'):
        raise ValueError('type of saving file must be pdf')

    # 检查pdfs参数
    if isinstance(pdfs, list):
        for item in pdfs:
            if not FileName(item).is_ext('.pdf'):
                raise ValueError('\'{}\' is not pdf'.format(item))
    else:
        raise TypeError('param pdfs must be list')

    # 合并pdf
    merged_pdf = Pdf.new()
    for file in pdfs:
        src = Pdf.open(file)
        merged_pdf.pages.extend(src.pages)
    pdf.make_dir()
    merged_pdf.save(pdf_path)

    return merged_pdf



import os
import sys
from xyw_utils.errors import NoFileError


def get_filepath():
    """
    获取当前运行文件的绝对地址
    :return:
    """
    # 此处排除了命令行和pycharm中console这两种情况
    if sys.argv[0] == '' \
            or os.path.split(sys.argv[0])[-1] == 'pydevconsole.py':
        raise NoFileError('must be used in file')
    return os.path.abspath(sys.argv[0])


def get_filedir():
    """
    获取当前运行文件所在文件夹的绝对地址
    :return:
    """
    return os.path.dirname(get_filepath())


def change_to_maindir():
    """
    切换工作目录为当前运行文件所在文件夹，通常用在main函数开头
    :return:
    """
    os.chdir(get_filedir())


class FileName:
    """
    文件名类，将完整地文件路径拆分为路径、文件名、文件后缀三部分存储
    """
    def __init__(self, fullpath):
        self.path, filename = os.path.split(fullpath)
        self.name, ext = os.path.splitext(filename)
        self.__ext = ext.lower()

    @property
    def ext(self):
        return self.__ext

    @ext.setter
    def ext(self, ext):
        if not(ext == '' or (len(ext) > 1 and ext[0] == '.')):
            raise ValueError('param ext must be \'\' or \'.*\'')
        self.__ext = ext.lower()

    def is_ext(self, ext):
        """
        判断后缀是否相同
        :param ext:
        :return:
        """
        if isinstance(ext, str):
            return self.ext == ext
        elif isinstance(ext, list):
            return self.ext in ext
        else:
            raise TypeError('param ext must be string or list')

    def is_image(self):
        """
        判断是否为图片文件
        :return:
        """
        return self.is_ext(['.jpg', '.jpeg', '.png', '.gif'])

    def is_file_exist(self):
        """
        判断文件是否存在
        :return:
        """
        if os.path.isdir(self.fullpath):
            raise RuntimeError('the path is a dir')
        return os.path.exists(self.fullpath)

    def is_dir_exist(self):
        """
        判断文件所在文件夹是否存在
        :return:
        """
        if os.path.isfile(self.path):
            raise RuntimeError('the path is a file')
        return os.path.exists(self.path) if self.path != '' else True

    def make_dir(self):
        """
        创建文件所在文件夹
        :return:
        """
        if not self.is_dir_exist():
            os.makedirs(self.path)

    def set_fullpath(self, fullpath):
        """
        设置完整路径名
        :param fullpath:
        :return:
        """
        self.path, filename = os.path.split(fullpath)
        self.name, self.ext = os.path.splitext(filename)

    def get_filename(self):
        """
        获取文件名
        :return:
        """
        return ''.join([self.name, self.ext])

    def set_filename(self, filename):
        """
        设置文件名
        :param filename:
        :return:
        """
        self.name, self.ext = os.path.splitext(filename)

    def get_fullpath(self):
        """
        设置完整路径名
        :return:
        """
        return os.path.join(self.path, self.get_filename())

    def __eq__(self, other):
        return self.path == other.path and self.name == other.name and self.ext == other.ext

    def __repr__(self):
        return 'FileName(r\'%s\')' % self.fullpath

    filename = property(get_filename, set_filename)
    fullpath = property(get_fullpath, set_fullpath)


if __name__ == '__main__':
    f = FileName(r'C:\Users\Administrator\Desktop\xyw\xyw_test\d.py\xue.JPG')
    print(f.ext)
    print(f.filename)
    print(f.is_file_exist())
    f.filename = 'xue.jpg'
    print(f.fullpath)
    print(f.is_file_exist())
    # f.make_dir()
    print(f)
    # print(eval(r"FileName(r'C:\Users\Administrator\Desktop\xyw\xyw_test\d.py\xue.jpg')"))
    print(f == eval(repr(f)))
    f.ext = '.PnG'
    print(f)



import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='xyw-utils',
    version='0.0.10',
    author='二炜',
    author_email='1174543101@qq.com',
    description='个人工具合集',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    # package_data={
    #     'examples': ['*.csv', '*.xlsx', '*.exe'],
    #     'example/imgs': ['*.jpg']
    # },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pandas',
        'numpy',
        'scipy',
        'pygal',
        'cairosvg',
        'cssselect2',
        'pikepdf',
        'img2pdf',
        'openpyxl',
        'xlwt',
        'xlsxwriter',
        'matplotlib',
    ],
)
